#!/usr/bin/perl

use strict;
use warnings;

eval { require Module::Build; } or do {
    die "error: Missing Module::Build module, cannot proceed.\n";
};

my $build = Module::Build->new(
    dist_name     => 'dh-golang',
    dist_abstract => 'debhelper golang integration',
    dist_version  => '1.0',
    license       => 'GPL_2',

    module_name => 'Debian::Debhelper::Buildsystem::golang',

    configure_requires => {
        'Module::Build' => 0,
    },
    requires => {
        'perl' => '5.36.0',
        'Dpkg' => 0,
    },
);

$build->create_build_script();
